import { Interface } from "readline"
import { Observer, Subject, Subscribable, Unsubscribable } from "rxjs"
/*Fis actor to declare its own subscribable but extended from RXJS */


/*  ===========Message Schema Mode =============

    Following intefaces pertain to messaging model.
    Declared as interfaces so that new member elements can be 
    added. This definition only list type of messages recognised 
    by the base Actor. Messages streamed from subscribables if 
    necessary need to be mapped to message type recognised by 
    the actor.
    
    The concrete message will need to add type
    of schema corresponding to each of the message type in the 
    message model (refer to src.common.messageModel.messageSchema)
    
    File location: src.messageModel.messageType folder.
    
    */


export interface MessageSchemaService extends CRUD {
    /*  collection of method signatures*/
}

export interface MessageSchema {
    schemaId:       string // identifier 
    schemaName:     string // unique legible name
    schemaDef:      string // jason description of schema
}

export interface MessageNameSchemaMap {
/* mapping message name to message schema by 
    message domain */
    messageName:    string
    schemaId:       string
    messageDomain:  string
    /*  a same message name can be used in different domain
        but with different schema.  */
}
export class MessageNameSchemMapImpl implements MessageNameSchemaMap {
    messageName!:    string
    schemaId!:       string
    messageDomain!:  string //"DEFAULT"  "ACTOR" "CQRSACTOR"
    constructor (schemaId: string, messageDomain: string) {
        /*  all constructor parameter values to be provided upon creation. 
            Class unique by all fields */

    } 
}

export interface MessageSchemaModel {
/*  Descsription steriotype for all Fis messages.
    Has reference to message schema described as interfaces.
    MessageSchema is a network structure (not a tree hierarchy) where
    a child schema may have multiple parent schema
    Definition begins at the lowest denomiator 
    schema, or lowest node.
    On handling version of message schema, refer to version model
*/
    messageSchema: MessageSchema
    messageSchemaService: MessageSchemaService
}

export class MessageSchemaComposite {
/*  Associated nodes of a composite schema.*/
    messageSchema: MessageSchema
    compositeEntries: Array <MessageSchema>[]
    
    constructor (messageSchema: MessageSchema, compositeEntries: Array <MessageSchema>[]) {
        this.messageSchema = messageSchema 
        this.compositeEntries = compositeEntries
    }

    /* validate:
         - Note association is one directional
         - duplicate association not allowed
         - messageSchema is the container of composite entries */
}


export class MessageSchemaGrouping {}
/* multi-level grouping message schema type */ 




export class MessageSchemaModelImpl implements MessageSchemaModel {

    messageSchema: MessageSchema
    messageSchemaService: MessageSchemaService
    
    constructor (messageSchema: MessageSchema,
        messageSchemaService: MessageSchemaService
    ) {
        this.messageSchema = messageSchema
        this.messageSchemaService = messageSchemaService
    }
    /*  base validation:
            - if schemaId is not provided by creator, then self generate 
            - schemaDef should be a valid schema (call api to check)
            - value changed should emit value changed event to be
            - handled by for example version logging

        element validation:
            - schema definition should not contain other schema
    */
}


/*  ======= the following section to be filed as:
    src/common/grouping/ ============ 
    
    This section to be completed ....
    
    */

export class GroupDataModel {
/*  Grouping of any entity. */
    groupId!     : string // unique value
    groupName!   : string // unique value
    groupShortName!  : string // can be duplicated
}
        
export class GroupChildDataModel extends GroupDataModel{
    groupChildList!    : Array<GroupDataModel>[]
}
          

/*  ======= this section of the code to be saved in file 
    src/common/services/ ========*/

export interface CRUD {
    /*  standard method signature for common entity behaviour*/
        create<T>(dataModel: T): null
        read<T>(): T
        update<T>(dataModel: T): null
        delete<T>(): null
}
    
/* this section to be saved to src/common/exceptions/ */
type ErrorSeverity = "RECOVERABLE" | "IRRECOVERABLE" 

export class FisErrorDataModel {
    errorSeverity!   : "RECOVERABLE" | "IRRECOVERABLE"
    errorMessage!    : string
    errorSourceId!   : string
    errorSource!     : string
}

export class FisError extends Error {
/*  extend error class to track error details */

    errorDataModel   : FisErrorDataModel
    constructor (errorDataModel : FisErrorDataModel) {
        super()
        this.errorDataModel = errorDataModel}
}



/*  ======= 
    Message names recognized by Actor. Names are defined
    as object constants. Can be extended via object union.

       
    file: ..src/common/fisactor/messages ========*/

/* Actor finite state */
export const ActorFiniteState = { 
/*  Actor finite state. Refer to Actor FSM doc. 
    For CQRS Actor which is an 'extension' to Actor,
    only one command is executed at any one time.*/
    Initialising: "INITIALISING", 
    Ready: "READY", 
    Executing: "EXECUTING", 
    Paused: "PAUSED", 
    ShuttingDown: "SHUTTINGDOWN", 
    Terminated: "TERMINATED" 
} as const


/** Below section: Define Actor message names (enumerators) */
export const CommandsTemp = {
    Initialise: "INITIALISE", // initialise after created
    Start:  "START", // start listening and responding
    Subscribe:  "SUBSCRIBE", // subscribe a subscribable
    UnSubscribe: "UNSUBSCRIBE", // unsubscribe a subscribable
    Pause:  "PAUSE", // pause further execution 
    Reset:  "RESET", // reset to initial state
    ShutDown: "SHUTDOWN", // shutdown
 } as const

//////

export type ActorMessageName = string

export interface ActorMessages <T> {
    // command requests
    Initialise: T       // initialise after created
    Start:  T           // start listening and responding
    UnSubscribe: T      // unsubscribe a subscribable
    Pause:  T           // pause further execution 
    Reset:  T           // reset to initial state
    ShutDown: T         // shutdown   

    // query requests
    GetActorIdentifier: T
    GetActorState:  T
    GetSubscribables: T // return list of subscribable identities
    GetSubscribabers: T, // return list of observer identities

    // query responses
    QueryData: T
    QuerySummary: T // return list of subscribable identities
    QueryStatus: T, // return list of observer identities

    // subscription requests
    Subscribe:  T
    
    // subscription responses
    subscriptionData:   T
    subscriptionSummary: T
    subscriptionStatus: T

    // general request status
    RequestReceived: T // received and que
    RequestCancelled: T // cancelled from que
    RequestExeStarted: T
    RequestExeEnded: T
    RequestCompleted: T
    RequestUnRecognised: T
    RequestExecutionError: T

    // notification: irrecoverable error
    InternalError: T // within actor
    EnvironmentError: T // due to external factor
    UnknownError: T // unable to identify source
    
}

export class ActorMessagesNameImpl <MessageName> implements ActorMessages <string> {
    Initialise = "INITIALISE"       // initialise after created
    Start = "START"           // start listening and responding
    UnSubscribe = "UNSUBSCRIBE"     // unsubscribe a subscribable
    Pause = "PAUSE"           // pause further execution 
    Reset = "RESET"       // reset to initial state
    ShutDown = "SHUTDOWN"      // shutdown   

    // query requests
    GetActorIdentifier = "GETACTORIDENTIFIER"
    GetActorState = "GETACTORSTATE"
    GetSubscribables = "GETSUBSCRIBABLES" // return list of subscribable identities
    GetSubscribabers = "GETSUBSCRIBERS" // return list of observer identities

    // query responses
    QueryData = "QUERYDATA"
    QuerySummary = "QUERYSUMMARY" // return list of subscribable identities
    QueryStatus = "QUERYSTATUS" // return list of observer identities

    // subscription requests
    Subscribe = "SUBSCRIBE"
    
    // subscription responses
    subscriptionData = "SUBSCRIPTIONDATA"
    subscriptionSummary = "SUBSCRIPTIONSUMMARY"
    subscriptionStatus = "SUBSCRIPTIONSTATUS"

    // general request status
    RequestReceived = "REQUESTRECEIVED" // received and que
    RequestCancelled = "REQUESTCANCELLED" // cancelled from que
    RequestExeStarted = "REQUESTEXESTARTED"
    RequestExeEnded = "REQUESTEXEENDED"
    RequestExeSuccessful = "REQUESTEXESUCCESSFUL"
    RequestUnRecognised = "REQUESTUNRECOGNISED"
    RequestExeError = "REQUESTEXEERROR"

    // notification: irrecoverable error
    InternalError = "INTERNALERROR" // within actor
    EnvironmentError = "ENVIRONMENTERROR" // due to external factor
    UnknownError = "UNKNOWNERROR" // unable to identify source

}

export interface ActorCommands <T>  {
    Initialise: T     // initialise after created
    Start:  T         // start listening and responding
    Subscribe:  T     // subscribe a subscribable
    UnSubscribe: T  // unsubscribe a subscribable
    Pause:  T             // pause further execution 
    Reset:  T             // reset to initial state
    ShutDown: T        // shutdown
    
 }

/** Below section: Define Actor message names (enumerators) */
export class ActorCommandNamesImpl <MessageName> implements ActorCommands <string> {
    Initialise = "INITIALISE" // initialise after created
    Start = "START" // start listening and responding
    Subscribe = "SUBSCRIBE" // subscribe a subscribable
    UnSubscribe = "UNSUBSCRIBE" // unsubscribe a subscribable
    Pause = "PAUSE" // pause further execution 
    Reset = "RESET" // reset to initial state
    ShutDown = "SHUTDOWN" // shutdown
 }

//////

/** Below section: Define Actor message names (enumerators) */
export const Commands = {
    Initialise: "INITIALISE", // initialise after created
    Start:  "START", // start listening and responding
    Subscribe:  "SUBSCRIBE", // subscribe a subscribable
    UnSubscribe: "UNSUBSCRIBE", // unsubscribe a subscribable
    Pause:  "PAUSE", // pause further execution 
    Reset:  "RESET", // reset to initial state
    ShutDown: "SHUTDOWN", // shutdown
 } as const

export const CommandResponses = {
    CommandReceived: "COMMANDRECEIVED", // received and que
    CommandCancelled: "COMMANDCANCELLED", // cancelled from que
    CommandExeStarted: "COMMANDEXESTARTED",
    CommandExeEnded: "COMMANDEXEENDED",
    CommandCompleted: "COMMANDCOMPLETED",
    UnRecognisedCommand: "UNRECOGNISEDCOMMAND",
    CommandExecutionError: "COMMANDEXECUTIONERROR"
}   as const

export const Queries = {
    
    GetActorIdentifier: "GETACTORIDENTY",
    GetSubscribables: "GETSUBSCRIBABLES", // return list of subscribable identities
    GetSubscribabers: "GETSUBSCRIBERS", // return list of observer identities
}   as const

export const QueryResponses = {
    QueryData: "QUERYDATA",
    QuerySummary: "QUERYSUMMARY", // return list of subscribable identities
    QueryStatus: "QUERYSTATUS", // return list of observer identities
}   as const

export const Subscriptions = { 
    Subscribe:  "SUBSCRIBE"
}   as const

export const SubscriptionResponses = {
    subscriptionData:   "SUBSCRIPTIONDATA",
    subscriptionSummary: "SUBSCRIPTIONSUMMARY",
    subscriptionStatus: {
        Completed: "COMPLETED",
        Error: "ERROR"                        
    }
} as const

export const IrrecoverableError = {
    InternalError: "INTERNAL ERROR", // withinn actor
    EnvironmentError: "ENVIRONMENTERROR", // outside actor
    UnknownError: "UNKNOWNERROR" // unable to identify
} as const

export const StateChanged = "STATECHANGED" as const

export const RequestMessages = {
    Commands,
    Queries,
    Subscriptions
}

/** Following Section: Fis message schema */

export interface FisMessageSchema {
    /** Standard Fis message schema definition */
    schemaId:       string
    messageName:    string
}

/** Following Section: Actor message schema */

export interface ActorMessageSchema {
/*  schema definition of an Actor message.
    T correspond to a message group such as ActorCommands */

    messageType!:   T // e.g. "SUBSCRIBE", "GETACTORIDENTITY"
    schemaId!:      S

    constructor (messageType: T, messageSchema: S) {}
}



/* ======= file: ..src/common/fisactor/services ============*/

export class ActorCommandServicesImpl { 
/*  Stateful actor service to be instantiated by concrete class.
    To add in behaviour and model state repository */
    
    context :ActorServicesContext
    // commandService = this.context

    constructor (context: ActorServicesContext) {
        this.context = context
    }
    
    executeCommand (command: string) {
        /* 
         */
        return null}
}
    

interface ActorQueryServices { 
/*  Stateful actor service to be instantiated by concrete class.
    To add in behaviour and model state repository */
}

interface ActorSubscriptionServices { 
/*  Stateful actor service to be instantiated by concrete class.
    To add in behaviour and model state repository */
}


interface ActorContext<T> { 
/*  Environment within which actor operates. Provided by creator in 
    constructor */
    defaultSubscribers: FisActor<T>[]
    /* instances of observer actors expected to subscribe upon creation of this actor */
    defaultSubscribables: FisActor<T>[]
    /*  instances of actors to auto subscribe to upon creation.
        This would include the Supervisor */
//    actorMessages: 
}

export interface ActorServicesContext { 
/* environment within which the actor service operates */
}

// Please note members of an interface are ALWAYS PUBLIC.
interface FisActor<T> extends Subscribable<T> {
    subscribe(observer: Partial<Observer<T>>): Unsubscribable
}


/*  ============ Actor model base ================== */


export class FisActorImpl<T> implements FisActor<T> {

    /*  A Fis actor is stateful object which implements Subscribable 
    interface. It can be anything including:
        business domain object
        UI component
        message forwarder
        actor directory receptionist
        actor container
        remote transmission service 
        mircroservice
        client proxy
        domain proxy
        backoffice App Manager

    Behavior of an actor is solely dependent on the actor service
    passed from its creator actor
        
    Folder location: src.common.actor
*/

    protected actorId!: String
    /* autogenerate a GUID value upon creation */

    protected actorName: String = "No Name" //might be reassigned during run time
    protected incomingBus: Subscribable<T> = new Subject
    protected outgoingBus: Subscribable<T> = new Subject
    protected actorService!: ActorService // service provided by creator 
    protected actorContext!: ActorContext<T> // context provided by creator
    protected actorServiceContext!: ActorServiceContext
    protected supervisorActor!: FisActor<T> 
    /*  Supervisor assigned by creator. A creator us usually but not aways the Supervisor.
        An Actor will always has a Supervisor Actor unless it is a Root Actor.
        Note that an Aggregate Root Actor (for CQRS context) is not the same as Root Actor
    */

    
    constructor(actorContext: ActorContext<T>) {
        // logic here
    }

    public subscribe(observer: Partial<Observer<T>>): Unsubscribable {
        /* to subscribe from output bus for now 
        */

        throw new Error("Method to be defined....")
    }


}

interface Actor<T> extends Subscribable<T> {
    actorId: String
    actorName: String
    actorStream: Subscribable<T>
}

abstract class Sample<T> implements Actor<T> {
    actorId!: String
    actorName!: String
    actorStream!: Subscribable<T>

    subscribe(observer: Partial<Observer<T>>): Unsubscribable {
        throw new Error("Method not implemented.")
    }

}